# pyawd - AcousticWaveDataset
# Tribel Pascal - pascal.tribel@ulb.be
from typing import Tuple, List, Union

import numpy as np
import devito as dvt
import matplotlib.pyplot as plt
from matplotlib.colors import TABLEAU_COLORS
import torch

from tqdm.auto import tqdm
from pyawd import VectorAcousticWaveDataset, VelocityModel2D
from pyawd.GenerateVideo import generate_quiver_video, generate_2d_density_video
from pyawd.utils import create_explosive_source
from pyawd.Marmousi import Marmousi

COLORS = TABLEAU_COLORS
dvt.configuration['log-level'] = "WARNING"



class VectorAcousticWaveDataset2D(VectorAcousticWaveDataset):
    """
    A Pytorch dataset containing acoustic waves propagating in the Marmousi velocity field
    """
    def __init__(self, size: int, dx: float = 1000/32., nx: int = 32, sx: float = 1., ddt: float = 0.01,
                  dt: float = 2, t: float = 10, interrogators: List[Tuple] = None, velocity_model: Union[
                  str, float] = 300., attenuation_factor: float = 0.5, openmp: bool = False):
        """
        Args:
            size (int): The number of samples to generate in the dataset
            dx (float): The discretisation rate of the array
            nx (int): The discretisation size of the array
            sx (float): The sub-scaling factor of the array (0.5 means 1/2 values are returned)
            ddt (float): The time step used for the Operator solving iterations
            dt (float): The time step used for storing the wave propagation step (this should be higher than ddt)
            t (float): The simulations duration
            velocity_model (str | float): either:
                - A string identifier specifying a velocity framework
                - A float, specifying a constant wave propagation speed. Currently, only this type can be used with dim=3
            attenuation_factor (float): The attenuation factor in the acoustic wave equation
            openmp (bool): Use openmp optimization
        """
        if interrogators is None:
            interrogators = [tuple(0 for _ in range(2))]
        super().__init__(size=size, dx=dx, nx=nx, sx=sx, ddt=ddt, dt=dt, t=t, interrogators=interrogators,
                         attenuation_factor=attenuation_factor, openmp=openmp)

        self.grid = dvt.Grid(shape=tuple(self.nx for _ in range(2)), extent=tuple(self.dx*self.nx for _ in range(2)))
        self._u = dvt.VectorTimeFunction(name='u', grid=self.grid, space_order=2, save=self.ndt, time_order=2)
        self._f = dvt.VectorTimeFunction(name='f', grid=self.grid, space_order=1, save=self.ndt, time_order=1)
        self.velocity_model = dvt.Function(name='c', grid=self.grid)
        if velocity_model == "Marmousi":
            self._velocity_model = Marmousi(self.nx)
            self._display_velocity_model = True
            self.velocity_model.data[:] = self._velocity_model.get_data()
            self.max_velocities = ((np.random.random(size=size))+0.5) * 600
        elif isinstance(velocity_model, float) or isinstance(velocity_model, int):
            self._velocity_model = VelocityModel2D(self.nx)
            self._velocity_model.set_value(velocity_model)
            self._display_velocity_model = False
            self.velocity_model.data[:] = self._velocity_model.get_data()
            self.max_velocities = ((np.random.random(size=size))+0.5)*0.1
        self.epicenters = np.random.randint(-self.nx // 2, self.nx // 2, size=(self.size, 2)).reshape(
            (self.size, 2))

    def solve_pde(self, idx: int) -> np.ndarray:
        """
        Solves the Acoustic Wave Equation for the idx parameters.
        Returns:
            (numpy.ndarray): A numpy array containing the solutions for the `ndt` steps
        """
        self._u[0].data[:] = 1e-5 * (np.random.random(self._u[0].data[:].shape) - 0.5)
        self._u[1].data[:] = 1e-5 * (np.random.random(self._u[1].data[:].shape) - 0.5)
        s_t = self.amplitude_factor[idx] * np.exp(-self.ddt * (np.arange(self.ndt) - (self.force_delay[idx] / self.ddt)) ** 2)
        s_x, s_y = create_explosive_source(self.nx, x0=int(self.epicenters[idx][0]), y0=int(self.epicenters[idx][1]))
        self._f[0].data[:] = (np.tile(s_x, (s_t.shape[0], 1, 1)) * s_t[:, None, None])
        self._f[1].data[:] = (np.tile(s_y, (s_t.shape[0], 1, 1)) * s_t[:, None, None])
        op = dvt.Operator(dvt.Eq(self._u.forward,
                                 dvt.solve(dvt.Eq(self._u.dt2, self._f +
                                                  ((self.max_velocities[idx] * self.velocity_model) ** 2)
                                                  * self._u.laplace - self.attenuation_factor*self._u.dt),
                                           self._u.forward)), opt=('advanced', {'openmp': self.openmp}))
        op.apply(dt=self.ddt)
        return np.array([self._u[i].data for i in range(self._u.shape[0])])

    def plot_item(self, idx: int):
        """
        Plots the simulation of the $idx^{th}$ sample
        Args:
            idx (int): The number of the sample to plot
        """
        colors = {}
        i = 0
        for interrogator in self.interrogators:
            colors[interrogator] = list(COLORS.values())[i]
            i += 1
        item, _ = self[idx]
        epicenter = self.epicenters[idx]
        max_velocity = self.max_velocities[idx]
        f_delay = self.force_delay[idx]
        amplitude_factor = self.amplitude_factor[idx]
        fig = plt.figure(figsize=(self.nt * 5, 5))
        ax = []
        a, b = np.meshgrid(np.arange(self.nx), np.arange(self.nx))
        for i in range(self.nt):
            ax.append(fig.add_subplot(1, self.nt, i+1))
            if self._display_velocity_model:
                im = ax[i].imshow(self.velocity_model.data[::int(1 / self.sx), ::int(1 / self.sx)],
                            vmin=np.min(self.velocity_model.data[::int(1 / self.sx), ::int(1 / self.sx)]),
                            vmax=np.max(self.velocity_model.data[::int(1 / self.sx), ::int(1 / self.sx)]), cmap="gray")
                plt.colorbar(im, ax=ax[i])
                ax[i].quiver(a, b, item[0, i * (item.shape[1] // self.nt)], -item[1, i * (item.shape[1] // self.nt)],
                            scale=0.25)
            else:
                ax[i].quiver(a, b, item[0, i * (item.shape[1] // self.nt)],
                             item[1, i * (item.shape[1] // self.nt)],
                             scale=0.25)
            for interrogator in self.interrogators:
                ax[i].scatter(interrogator[0] + (self.nx // 2), interrogator[1] + (self.nx // 2), marker="1",
                              color=colors[interrogator], s=80)
            ax[i].set_title("t = " + str(i * (item.shape[1] // self.nt) * self.dt) + "s, \nVelocity factor = " +
                            str(max_velocity)[:5] + ", \nForce delay = " + str(f_delay)[:4] +
                            ", \nAmplitude factor = " + str(amplitude_factor)[:4] + "\nEpicenter = " + str(epicenter))
            ax[i].axis("off")
        plt.tight_layout()
        plt.show()

    def plot_interrogators_response(self, idx: int):
        """
        Plots the measurements taken by the interrogators for the $idx^{th}$ sample.
        Args:
            idx (int): The number of the sample to plot
        """
        colors = {}
        i = 0
        for interrogator in self.interrogators:
            colors[interrogator] = list(COLORS.values())[i]
            i += 1
        fig, ax = plt.subplots(ncols=len(self.interrogators), figsize=(len(self.interrogators) * 5, 5))
        y_lims = []
        _, full_data = self[idx]
        for i in range(len(self.interrogators)):
            data = full_data[self.interrogators[i]]
            y_lims += [np.min(data), np.max(data)]
            for j in range(data.shape[0]):
                if len(self.interrogators) == 1:
                    ax.plot(np.arange(0, self.ndt * self.ddt, self.ddt), data[j], linestyle=['-', '--'][j],
                            color=colors[self.interrogators[i]])
                else:
                    ax[i].plot(np.arange(0, self.ndt * self.ddt, self.ddt), data[j], linestyle=['-', '--'][j],
                               color=colors[self.interrogators[i]])
            if len(self.interrogators) == 1:
                ax.legend(["Abscissa", "Ordinate"])
                ax.set_title(str(self.interrogators[i]))
                ax.set_xlabel("time (s)")
                ax.set_ylabel("Amplitude")
                ax.set_ylim([np.min(data), np.max(data)])
                ax.grid("on")
            else:
                ax[i].legend(["Abscissa", "Ordinate"])
                ax[i].set_title(str(self.interrogators[i]))
                ax[i].set_xlabel("time (s)")
                ax[i].set_ylabel("Amplitude")
                ax[i].grid("on")
        if len(self.interrogators) > 1:
            for i in range(len(self.interrogators)):
                ax[i].set_ylim([np.min(y_lims), np.max(y_lims)])
            fig.suptitle("Velocity factor = " + str(self.max_velocities[idx])[:5] + "\nForce delay = " + str(
                self.force_delay[idx])[:4] + "\nAmplitude factor = " + str(self.amplitude_factor[idx])[:4] +
                         "\nEpicenter = " + str(self.epicenters[idx]))
        plt.tight_layout()
        plt.show()

    def generate_video(self, idx: int, filename: str, nb_images: int):
        """
        Generates a video representing the simulation of the $idx^{th}$ sample propagation
        Arguments:
            idx (int): the number of the sample to simulate in the video
            filename (str): the name of the video output file (without extension)
                        The video will be stored in a file called `filename`.mp4
            nb_images (int): the number of frames used to generate the video. This should be an entire divider of the number
                         of points computed when applying the solving operator
        """
        u = self.solve_pde(idx)
        generate_quiver_video(u[0][::self.ndt // nb_images], u[1][::self.ndt // nb_images], self.interrogators,
                              {i: u[:, ::self.ndt // nb_images, i[0] + (self.nx // 2), i[1] + (self.nx // 2)] for i in self.interrogators},
                              filename, nx=self.nx, dt=self.ndt * self.ddt / nb_images, c=self.velocity_model,
                              max_velocity=self.max_velocities[idx], display_velocity_model=self._display_velocity_model,
                              verbose=True)
    
    def generate_density_video(self, idx: int, filename: str, nb_images: int):
        """
        Generates a video representing the simulation of the $idx^{th}$ sample propagation, by length of the arrows
        Arguments:
            idx (int): the number of the sample to simulate in the video
            filename (str): the name of the video output file (without extension)
                        The video will be stored in a file called `filename`.mp4
            nb_images (int): the number of frames used to generate the video. This should be an entire divider of the number
                         of points computed when applying the solving operator
        """
        u = self.solve_pde(idx)
        generate_2d_density_video((u[0][::self.ndt // nb_images]**2+u[1][::self.ndt // nb_images]**2)**0.5, self.interrogators,
                              {i: u[:, ::self.ndt // nb_images, i[0] + (self.nx // 2), i[1] + (self.nx // 2)] for i in self.interrogators},
                              filename, nx=self.nx, dt=self.ndt * self.ddt / nb_images, c=self.velocity_model,
                              display_velocity_model=self._display_velocity_model, verbose=True)

    def __getitem__(self, idx) -> Tuple:
        """
        Returns:
            (Tuple): The epicenter, the simulation of the `idx`th sample, the maximal speed of propagation of the
             propagation field, the delay before the external force application, the force amplitude factor and
             the interrogated data
        """
        data = self.solve_pde(idx)
        return (torch.Tensor(data[:, ::int(self.ndt / self.nt),
                             ::int(1 / self.sx), ::int(1 / self.sx)]),
                {i: data[:, :, i[0] + (self.nx // 2), i[1] + (self.nx // 2)]
                 for i in self.interrogators})



# pyawd - AcousticWaveDataset
# Tribel Pascal - pascal.tribel@ulb.be
from typing import Tuple, List, Dict

import numpy as np
import devito as dvt

import torch.utils.data
import matplotlib.colors

COLORS = matplotlib.colors.TABLEAU_COLORS
dvt.configuration['log-level'] = "WARNING"


class AcousticWaveDataset(torch.utils.data.Dataset):
    """
    A Pytorch dataset containing acoustic waves propagating in the Marmousi velocity field
    """
    size: int
    """The number of samples to generate in the dataset"""
    nx: int
    """The discretisation size of the array (maximum size is currently 955)"""
    sx: float
    """The sub-scaling factor of the array (0.5 means $\\frac{1}{2}$ values are returned)"""
    dim: int
    """The number of dimensions of the simulations (2 or 3)"""
    ddt: float
    """The time step used for the Operator solving iterations"""
    dt: float
    """The time step used for storing the wave propagation step (this should be higher than ddt)"""
    ndt: int
    """The number of steps in the simulation, accessible for the interrogators"""
    t: float
    """The simulations duration"""
    nt: int
    """The number of steps in the simulations, for which the whole simulation is accessible"""
    interrogators: List[Tuple]
    """A list containing the coordinates of each interrogator"""
    interrogators_data: Dict[Tuple, List[torch.Tensor]]
    """The measurements of each interrogator"""
    attenuation_factor: float
    """The attenuation factor in the acoustic wave equation"""
    max_velocities: np.ndarray
    """The maximal velocity in the idx propagation field"""
    epicenters: np.ndarray
    """The epicenter of each simulation"""
    force_delay: np.ndarray
    """The delay of apparition of the external force for each simulation"""
    amplitude_factor: np.ndarray
    """The amplitude factor to multiply the external force with"""
    data: torch.Tensor
    """The simulations data"""
    openmp: bool
    """Use openmp optimization"""

    def __init__(self, size: int, dx: float = 1000/128., nx: int = 128, sx: float = 1., ddt: float = 0.01,
                 dt: float = 2, t: float = 10, interrogators: List[Tuple] = None, attenuation_factor: float = 0.5,
                 openmp: bool = False):
        """
        Args:
            size (int): The number of samples to generate in the dataset
            dx (float): The discretisation rate of the array
            nx (int): The discretisation size of the array
            sx (float): The sub-scaling factor of the array (0.5 means 1/2 values are returned)
            ddt (float): The time step used for the Operator solving iterations
            dt (float): The time step used for storing the wave propagation step (this should be higher than ddt)
            t (float): The simulations duration
            attenuation_factor (float): The attenuation factor in the acoustic wave equation
            openmp (bool): Use openmp optimization
        """
        try:
            if dt < ddt:
                raise ValueError('dt should be >= ddt')
            self.size = size
            self.dx = dx
            self.nx = nx
            self.sx = sx
            self.ddt = ddt
            self.dt = dt
            self.nt = int(t / self.dt)
            self.ndt = int(self.nt * (self.dt / self.ddt))
            self.interrogators = interrogators
            self.attenuation_factor = attenuation_factor
            self.force_delay = np.random.random(size) * (t/2)
            self.amplitude_factor = (0.5 * np.random.random(size) + 0.25) * 2
            self.data = torch.Tensor([])
            self.openmp = openmp

        except ValueError as err:
            print(err)

    def generate_data(self):
        """
        Generates the dataset content by solving the Acoustic Wave PDE for each of the `epicenters`
        """
        raise NotImplementedError('This class is abstract')

    def solve_pde(self, idx: int):
        """
        Solves the Acoustic Wave Equation for the idx parameters.
        Returns:
            (numpy.ndarray): A numpy array containing the solutions for the `ndt` steps
        """
        raise NotImplementedError('This class is abstract')

    def plot_item(self, idx: int):
        """
        Plots the simulation of the $idx^{th}$ sample
        Args:
            idx (int): The number of the sample to plot
        """
        raise NotImplementedError('This class is abstract')

    def plot_interrogators_response(self, idx: int):
        """
        Plots the measurements taken by the interrogators for the $idx^{th}$ sample.
        Args:
            idx (int): The number of the sample to plot
        """
        raise NotImplementedError('This class is abstract')

    def generate_video(self, idx: int, filename: str, nb_images: int):
        """
        Generates a video representing the simulation of the $idx^{th}$ sample propagation
        Arguments:
            idx (int): the number of the sample to simulate in the video
            filename (str): the name of the video output file (without extension)
                        The video will be stored in a file called `filename`.mp4
            nb_images (int): the number of frames used to generate the video. This should be an entire divider of the number
                         of points computed when applying the solving operator
        """
        raise NotImplementedError('This class is abstract')

    def set_scaling_factor(self, sx: float):
        """
        Fixes a new scaling factor (0.5 means $\\frac{1}{2}$ values are returned). It should be <= 1.
        Args:
            sx (float): the new scaling factor
        """
        if sx <= 1.:
            self.sx = sx
        else:
            print("The scaling factor should be lower or equal to 1.")

    def __len__(self):
        """
        Returns:
            (int): The number of simulations in the dataset
        """
        return self.size

    def __getitem__(self, idx):
        """
        Returns:
            (Tuple): The epicenter, the simulation of the `idx`th sample, the maximal speed of propagation of the
             propagation field, the delay before the external force application, the force amplitude factor and
             the interrogated data
        """
        raise NotImplementedError('This class is abstract')



# pyawd - VelocityModel2D
# Tribel Pascal - pascal.tribel@ulb.be
"""
Represents the velocity models as a numpy array
"""
import numpy as np
import matplotlib.pyplot as plt
from pyawd import VelocityModel


class VelocityModel2D(VelocityModel):
    """
    Represents the velocity models as a numpy array
    """
    nx: int = 32
    """
    The width of the field, in pixels
    """
    def __init__(self, nx: int = 32):
        """
        Args:
            nx (int): The width of the field, in pixels
        """
        super().__init__(nx=nx)
        self.data = np.ones((nx, nx))

    def get_data(self) -> np.ndarray:
        """
        Returns:
            - self.data: the velocity field
        """
        return self.data

    def plot(self):
        """
        Plots the field
        """
        plt.imshow(self.data)
        plt.colorbar()
        plt.tight_layout()



# pyawd - VelocityModel
# Tribel Pascal - pascal.tribel@ulb.be
"""
Represents the velocity models as a numpy array
"""
import numpy as np


class VelocityModel:
    """
    Represents the velocity models as a numpy array
    """
    nx: int = 32
    """
    The width of the field, in pixels
    """
    dim: int = 3
    """
    The number of dimensions of the model
    """
    def __init__(self, nx: int = 32):
        """
        Args:
            nx (int): The width of the field, in pixels
        """
        self.nx = nx
        self.data = np.array([])

    def get_data(self) -> np.ndarray:
        """
        Returns:
            - self.data: the velocity field
        """
        return self.data

    def set_value(self, value: float):
        """
        Initialize the data array with the specified value
        Args:
            - value (float): the value to initialize the array with
        """
        self.data[:] = value

    def plot(self):
        """
        Plots the field
        """
        raise NotImplementedError('This class is abstract')



# pyawd - AcousticWaveDataset
# Tribel Pascal - pascal.tribel@ulb.be
from typing import Tuple, List

import numpy as np
import devito as dvt
from matplotlib.colors import TABLEAU_COLORS

from pyawd import AcousticWaveDataset

COLORS = TABLEAU_COLORS
dvt.configuration['log-level'] = "WARNING"


class VectorAcousticWaveDataset(AcousticWaveDataset):
    """
    A Pytorch dataset containing acoustic waves propagating in the Marmousi velocity field
    """

    def __init__(self, size: int, dx: float = 1000 / 128., nx: int = 128, sx: float = 1., ddt: float = 0.01,
                 dt: float = 2, t: float = 10, interrogators: List[Tuple] = None, attenuation_factor: float = 0.5,
                 openmp: bool = False):
        """
        Args:
            size (int): The number of samples to generate in the dataset
            dx (float): The discretisation rate of the array
            nx (int): The discretisation size of the array
            sx (float): The sub-scaling factor of the array (0.5 means 1/2 values are returned)
            ddt (float): The time step used for the Operator solving iterations
            dt (float): The time step used for storing the wave propagation step (this should be higher than ddt)
            t (float): The simulations duration
            attenuation_factor (float): The attenuation factor in the acoustic wave equation
            openmp (bool): Use openmp optimization
        """
        super().__init__(size, dx, nx, sx, ddt, dt, t, interrogators, attenuation_factor, openmp=openmp)

    def save(self, filename: str):
        """
        Saves the dataset features to different files, starting with `filename`.
        This is intended to allow to retrieve the dataset by using "VectorAcousticWaveDataset2D().load(filename)`
        Args:
            filename (str): The base name of the files where the data will be saved
        """
        features = np.array([self.size, self.dx, self.nx, self.sx, self.ddt, self.dt, self.nt, self.ndt, self.attenuation_factor])
        np.save(filename+"_features.npy", features)
        np.save(filename+"_force_delays.npy", self.force_delay)
        np.save(filename+"_amplitude_factors.npy", self.amplitude_factor)
        np.save(filename+"_max_velocities.npy", self.max_velocities)
        np.save(filename+"_epicenters.npy", self.epicenters)
        np.save(filename+"_interrogators.npy", self.interrogators)
        np.save(filename+"_velocity_model.npy", self.velocity_model.data[:])

    @classmethod
    def load(cls, filename: str):
        """
        Loads a dataset that was saved through the `save()` method.
        Args:
            filename (str): The base name of the files where the data has been saved
        """
        features = np.load(filename+"_features.npy")
        dataset = cls(size=int(features[0]), dx=features[1], nx=int(features[2]), sx=features[3],
                      ddt=features[4], dt=features[5])
        dataset.nt = int(features[6])
        dataset.ndt = int(features[7])
        dataset.attenuation_factor = features[8]
        dataset.force_delay = np.load(filename+"_force_delays.npy")
        dataset.amplitude_factor = np.load(filename+"_amplitude_factors.npy")
        dataset.max_velocities = np.load(filename+"_max_velocities.npy")
        dataset.epicenters = np.load(filename+"_epicenters.npy")
        dataset.interrogators = [tuple(i) for i in np.load(filename+"_interrogators.npy")]
        dataset.velocity_model.data[:] = np.load(filename+"_velocity_model.npy")
        return dataset

    def generate_data(self):
        """
        Generates the dataset content by solving the Acoustic Wave PDE for each of the `epicenters`
        """
        raise NotImplementedError('This class is abstract')

    def solve_pde(self, idx: int):
        """
        Solves the Acoustic Wave Equation for the idx parameters.
        Returns:
            (numpy.ndarray): A numpy array containing the solutions for the `ndt` steps
        """
        raise NotImplementedError('This class is abstract')

    def plot_item(self, idx: int):
        """
        Plots the simulation of the $idx^{th}$ sample
        Args:
            idx (int): The number of the sample to plot
        """
        raise NotImplementedError('This class is abstract')

    def plot_interrogators_response(self, idx: int):
        """
        Plots the measurements taken by the interrogators for the $idx^{th}$ sample.
        Args:
            idx (int): The number of the sample to plot
        """
        raise NotImplementedError('This class is abstract')

    def generate_video(self, idx: int, filename: str, nb_images: int):
        """
        Generates a video representing the simulation of the $idx^{th}$ sample propagation
        Arguments:
            idx (int): the number of the sample to simulate in the video
            filename (str): the name of the video output file (without extension)
                        The video will be stored in a file called `filename`.mp4
            nb_images (int): the number of frames used to generate the video. This should be an entire divider of the number
                         of points computed when applying the solving operator
        """
        raise NotImplementedError('This class is abstract')

    def get_epicenter(self, idx):
        """
        Returns the epicenter of the $idx^{th}$ sample
        """
        return self.epicenters[idx]

    def get_max_velocity(self, idx):
        """
        Returns the maximal velocity of the $idx^{th}$ sample
        """
        return self.max_velocities[idx]

    def get_force_delay(self, idx):
        """
        Returns the delay before the external force occurring in the $idx^{th}$ sample
        """
        return self.force_delay[idx]

    def get_amplitude_factor(self, idx):
        """
        Returns the amplitude factor of the external force in the $idx^{th}$ sample
        """
        return self.amplitude_factor[idx]

    def __getitem__(self, idx):
        """
        Returns:
            (Tuple): The epicenter, the simulation of the `idx`th sample, the maximal speed of propagation of the
             propagation field, the delay before the external force application, the force amplitude factor and
             the interrogated data
        """
        raise NotImplementedError('This class is abstract')



# pyawd
# Tribel Pascal - pascal.tribel@ulb.be
r"""
<img src="https://github.com/pascaltribel/pyawd/raw/main/Logo.png?raw=true" alt="Logo" width="60%"/>

# pyawd - A PyTorch Acoustic Wave Dataset Generator
Pyawd (standing for Pytorch Acoustic Wave Dataset) is a powerful tool for building datasets containing custom simulations of the propagation of Acoustic Wave through a given medium.
It uses the finite differences scheme (implemented in the Devito package) to solve the Acoustic Wave Equation, and offers convenient tools for the customisation of the parameters, the handling of the data, the visualisation of the simulations.

Here is an example of the kind of visualisation PyAWD offers. The two first pannels show the propagation in 3 dimensions, and the three last ones show the measured movement at the three interrogators, placed at the cube top surface.

<video src="https://github.com/pascaltribel/pyawd/raw/main/examples/test3.mp4" width="1000" controls></video>


## Acoustic Wave Equation
The equation of propagation of an acoustic wave is given by 
$$\frac{d^2u}{dt^2} = c \nabla^2 u + f(x, y, t)$$ 
where
- $u(x, y, t)$ is the displacement field, and can be either a scalar or a vector field
- $c(x, y, t)$ is the wave  propagation speed
- $\nabla^2$ is the <i>laplacian operator</i>
- $f(x, y, t)$ is an external force applied on the system


## Installation
The package (along with the dependencies) is accessible via [PyPI](https://pypi.org/project/PyAWD/):

```bash
pip install pyawd
```

## Documentation
The API documentation is available [here](https://pascaltribel.github.io/pyawd/).
Basic help is provided for each class and function, and is accessible via the Python `help()` function.

## Getting started

Basic imports:
```python
from pyawd import *
```

Let us generate a Dataset made of 10 simulations. Each simulation is run in a $250\times 250$ matrix. We store the field state every $2$ seconds and we run the simulation for $10$ seconds:

```python
dataset = VectorAcousticWaveDataset2D(2, nx=128, dt=2, t=10, velocity_model="Marmousi")
dataset.max_velocities[0] = 500
```

Then we plot the first simulation. The &#128960; character shows the interrogator position:

```python
dataset.plot_item(0)
```

Which outputs the following figure:

<img src="https://github.com/pascaltribel/pyawd/raw/main/examples/example.png" alt="Example" width="60%"/>

By default, the point `(0, 0)` contains an interrogator. This means that the continuous measurement on this position (at least with a $\Delta t=ddt$) can be plot by:

```python
dataset.plot_interrogators_response(0)
```

<img src="https://github.com/pascaltribel/pyawd/raw/main/examples/interrogator_example.png" alt="Example" width="60%"/>

## More advanced usage
Using the `VectorAcousticWaveDataset3D` class allows producing simulations in 3D:

```python
dataset_3d = VectorAcousticWaveDataset3D(1, nx=32, t=10, interrogators=[(0, 10, 15), (0, -10, 15)], velocity_model=300.)
```

For visualisation, the method

```python
dataset_3d.generate_video(0, "VAWD3D", 300)
```

generates the following video:

<video src="https://github.com/pascaltribel/pyawd/raw/main/examples/VAWD3D.mp4" width="1000" controls></video>


## Examples
Multiple IPython notebooks are presented in the [examples](https://github.com/pascaltribel/pyawd/tree/main/examples) directory. If [Jupyter](https://jupyter.org) is installed, those examples can be explored by starting Jupyter:

```bash
jupyter-notebook
```

However, those are provided in HTML for your convenience. Just click on their name in the following list:

- [`ScalarAcousticWavePropagation.ipynb`](https://pascaltribel.github.io/pyawd/pyawd/ScalarAcousticWavePropagation.html): an introduction to PDE solving and simulation using Devito applied on the scalar acoustic wave propagation
- [`VectorAcousticWavePropagation.ipynb`](https://pascaltribel.github.io/pyawd/pyawd/VectorAcousticWavePropagation.html): an introduction to PDE solving and simulation using Devito applied on the vector acoustic wave propagation
- [`VectorAcousticWaveDataset.ipynb`](https://pascaltribel.github.io/pyawd/pyawd/VectorAcousticWaveDataset.html): an introduction to the VectorAcousticWaveDataset possibilities
- [`Marmousi.ipynb`](https://pascaltribel.github.io/pyawd/pyawd/Marmousi.html): a visualisation of the Marmousi velocity field used in the simulations
- [`Interrogators.ipynb`](https://pascaltribel.github.io/pyawd/pyawd/Interrogators.html): an introduction to the PyAWD Interrogators usage
- [`GenerateVectorAcousticWaveDataset.ipynb`](https://pascaltribel.github.io/pyawd/pyawd/GenerateVectorAcousticWaveDataset.html): how to generate dataset using `pyawd`
- [`SpatioTemporalVaryingWavePropagationSpeedField.ipynb`](https://pascaltribel.github.io/pyawd/pyawd/SpatioTemporalVaryingWavePropagationSpeedField.html): how to create a spatio-temporal varying propagation field
"""
from pyawd.VelocityModel import VelocityModel
from pyawd.VelocityModel2D import VelocityModel2D
from pyawd.VelocityModel3D import VelocityModel3D
from pyawd.Marmousi import Marmousi
from pyawd.AcousticWaveDataset import AcousticWaveDataset
from pyawd.VectorAcousticWaveDataset import VectorAcousticWaveDataset
from pyawd.VectorAcousticWaveDataset2D import VectorAcousticWaveDataset2D
from pyawd.VectorAcousticWaveDataset3D import VectorAcousticWaveDataset3D
